# nozzle.R - Report library for high-throughput data analysis pipelines.
#
# Author: Nils Gehlenborg / Harvard Medical School & Broad Institute
#
# This is a preview release. The library will be made available under an open source license.


# --- global variables ----------------------------------------------------------------------------

.nozzleEnv <- new.env();

.PACKAGE.VERSION <- "0.2.99";

.ELEMENT.REPORT <- "_report_";
.ELEMENT.SECTION <- "_section_";
.ELEMENT.SUBSECTION <- "_subsection_";
.ELEMENT.SUBSUBSECTION <- "_subsubsection_";
.ELEMENT.PARAGRAPH <- "_paragraph_";
.ELEMENT.LIST <- "_list_";
.ELEMENT.PARAMETERLIST <- "_list_";
.ELEMENT.TABLE <- "_table_";
.ELEMENT.FIGURE <- "_figure_";
.ELEMENT.CITATION <- "_citation_";
.ELEMENT.RESULT <- "_result_";

.SEVERITY.LOG <- 0;
.SEVERITY.WARNING <- 1;
.SEVERITY.ERROR <- 2;

# names for popout windows
.POPOUT.FIGURE.WINDOW <- "nozzle_popout";
.POPOUT.TABLE.WINDOW <- "nozzle_popout";

DEFAULT.REPORT.FILENAME <- "report";
DEFAULT.SIGNIFICANT.ENTITY <- "statistically significant findings";


# location relative to inst diretory of package
.NOZZLE.JAVASCRIPT.PATH <- "js";
.NOZZLE.JAVASCRIPT.FILE <- "nozzle.js";
.NOZZLE.JAVASCRIPT.MIN.FILE <- "nozzle.min.js";
.NOZZLE.JQUERY.MIN.FILE <- "jquery-1.5.1.min.js";
.NOZZLE.TABLESORTER.MIN.FILE <- "jquery.tablesorter.min.js";
	
.NOZZLE.CSS.PATH <- "css";
.NOZZLE.CSS.FILE <- "nozzle.css";
.NOZZLE.CSS.MIN.FILE <- "nozzle.min.css";

# formatting related defaults
TABLE.SIGNIFICANT.DIGITS <- 2;


# the ordering is important. It must be public < tcga < private.
PROTECTION.PUBLIC <- 0;
PROTECTION.TCGA <- 5;
PROTECTION.GROUP <- 5;
PROTECTION.PRIVATE <- 10;

IMAGE.TYPE.RASTER <- 0;
IMAGE.TYPE.SVG <- 1;
IMAGE.TYPE.PDF <- 2;


# output types
HTML.REPORT <- "html_standalone";
HTML.FRAGMENT <- "html_fragment";
RDATA.REPORT <- "rdata";


.REFERENCE.STRING <- "##REF##";

.onLoad <- function( libname, pkgname )
{
	.initEnvCounter( ".FIGURE.COUNTER" );
	.initEnvCounter( ".TABLE.COUNTER" );
	.initEnvCounter( ".FIGURE.SUPPLEMENTARY.COUNTER" );
	.initEnvCounter( ".TABLE.SUPPLEMENTARY.COUNTER" );
	.initEnvCounter( ".CITATION.COUNTER" );


	.initEnvCounter( ".ELEMENT.ID" );
	.initEnvCounter( ".RESULT.ID" );
	
	.setGlobalReportId( "" );
	
	# TODO: check if files exist
	assign( "cssCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.FILE), warn=FALSE ), env=.nozzleEnv );
	assign( "javaScriptCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JAVASCRIPT.FILE ), warn=FALSE ), env=.nozzleEnv );	
	assign( "cssMinCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );
	assign( "javaScriptMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JAVASCRIPT.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	

	assign( "jQueryMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JQUERY.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	
	assign( "jQueryTableSorterMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.TABLESORTER.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	

	assign( ".PACKAGE", pkgname, env=.nozzleEnv );
}



# --- utility functions ----------------------------------------------------------------------------

.updateEnvCounter <- function( counter, increment=1, environment=.nozzleEnv )
{
	assign( counter, get( counter, env=.nozzleEnv ) + increment, env=environment );
	
	return ( get( counter, env=environment )  );
}

.initEnvCounter <- function( counter, environment=.nozzleEnv )
{
	assign( counter, 0, env=environment );
}

.getNextElementId <- function()
{
	return ( .updateEnvCounter( ".ELEMENT.ID" ) );
}


.setGlobalReportId <- function( id, environment=.nozzleEnv )
{
	assign( ".REPORT.ID", id, env=environment );	
}


.getGlobalReportId <- function( environment=.nozzleEnv )
{
	return ( get( ".REPORT.ID", environment ) );	
}


.getPackageVersion <- function( environment=.nozzleEnv )
{
	return ( .concat( get( ".PACKAGE", environment ), "_", .PACKAGE.VERSION ) );	

	# set lib.loc!
	#return ( .concat( get( ".PACKAGE", environment ), " ", packageVersion( get( ".PACKAGE", environment ), lib.loc=??? ) ) );	
}


.initNumbers <- function()
{
	.initEnvCounter( ".FIGURE.COUNTER" );
	.initEnvCounter( ".TABLE.COUNTER" );
	.initEnvCounter( ".FIGURE.SUPPLEMENT.COUNTER" );
	.initEnvCounter( ".TABLE.SUPPLEMENT.COUNTER" );
	.initEnvCounter( ".CITATION.COUNTER" );
}


.getNextFigureNumber <- function()
{
	return ( .updateEnvCounter( ".FIGURE.COUNTER" ) );
}

.getNextTableNumber <- function()
{
	return ( .updateEnvCounter( ".TABLE.COUNTER" ) );
}

# supplementary figure numbers are negative
.getNextSupplementFigureNumber <- function()
{
	return ( .updateEnvCounter( ".FIGURE.SUPPLEMENT.COUNTER", increment=-1 ) );
}

# supplementary table numbers are negative
.getNextSupplementTableNumber <- function()
{
	return ( .updateEnvCounter( ".TABLE.SUPPLEMENT.COUNTER", increment=-1 ) );
}


.getNextCitationNumber <- function()
{
	return ( .updateEnvCounter( ".CITATION.COUNTER" ) );
}

.getNextResultId <- function()
{
	return ( .updateEnvCounter( ".RESULT.ID" ) );
}


.newReportId <- function()
{
	# inspired by http://www.broofa.com/Tools/Math.uuid.js
	# creates an almost RFC 4122-compliant version 4 UUID (see Note 1 below)
	# Note 1: since this UUID is going to be part of a JavaScript function name dashes are replaced with underscores
	# Note 2: this is written for convenience, not performance
	# Note 3: the "R" at position 1 is required to make the id a valid HTML id
	
	template <- unlist( strsplit( "Rxxxxxxxx_xxxx_4xxx_yxxx_xxxxxxxxxxxx", "" ) );
	hex <- unlist( strsplit( "0123456789ABCDEF", "" ) );
	
	result <- unlist( lapply( template, function( c ) {	
				  r <- floor( runif( n=1, min=1, max=16 ) );
				  v <- 0;
				  
				  if ( c == "x" )
				  {
					v <- r;
				  }
				  else if ( c == "y" )
				  {
				  	v <- ( r %% 4 ) + 8;
				  }
				  else
				  {
					return ( c );
				  }
				  
				  return ( hex[v] );
				}) );

	return ( paste( result, sep="", collapse="" ) );
}


.removeProtectedElements <- function( element, level )
{
	if ( element$protection > level )
	{
		return ( NULL );
	}
	
	if ( length( element$elements ) > 0 )
	{	
		retainedElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			result <- .removeProtectedElements( element$elements[[i]], level );
			
			if ( !is.null( result ) )
			{
				retainedElements[[length( retainedElements ) + 1]] <- result;
			}						
		}
		
		element$elements <- retainedElements;
	}
	
	return ( element );
}


# TODO: use separate counters for supplementary figures and tables
.updateNumberedElements <- function( element, supplement=FALSE )
{
	if ( element$type == .ELEMENT.FIGURE )
	{
		if ( !supplement )
		{
			element$number <- .getNextFigureNumber();
		}
		else
		{
			element$number <- .getNextSupplementFigureNumber();		
		}
	}

	if ( element$type == .ELEMENT.TABLE )
	{
		if ( !supplement )
		{
			element$number <- .getNextTableNumber();
		}
		else
		{
			element$number <- .getNextSupplementTableNumber();		
		}
	}

	if ( element$type == .ELEMENT.CITATION )
	{
		element$number <- .getNextCitationNumber();
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			if ( !supplement && ( element$type == .ELEMENT.RESULT ) )
			{
				supplement <- TRUE;
			}
			
			element$elements[[i]] <- .updateNumberedElements( element$elements[[i]], supplement=supplement );
		}
	}
	
	return ( element );
}


.countSignificantResults <- function( element )
{
	counter <- 0;
	
	if ( element$type == .ELEMENT.RESULT )
	{
		if ( element$isSignificant == TRUE )
		{
		  counter <- counter + 1;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			counter <- counter + .countSignificantResults( element$elements[[i]] );
		}
	}
	
	return ( counter );
}


.getReference <- function( referenceId, element )
{
	if ( element$id == referenceId )
	{
		if ( element$type == .ELEMENT.CITATION )
		{
			return ( .concat( "[", element$number, "]" ) );		
		}
		
		return ( .concat( element$niceType, " ", element$number ) );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			result <- .getReference( referenceId, element$elements[[i]] );
			
			if ( !is.null( result ) )
			{
				return ( result );
			}				
		}
	}
	
	return ( NULL );
}


.updateReferences <- function( element, report )
{
	if ( !is.null( element$text ) )
	{
		if ( is.list( element$text ) )
		{
			updatedText = "";
			
			if ( length( element$text ) > 0 )
			{
				for ( i in 1:length( element$text ) )
				{
					if ( is.list( element$text[[i]] ) )
					{
						next();
					}
					
					if ( substr( element$text[[i]], 0, nchar( .REFERENCE.STRING ) ) == .REFERENCE.STRING )
					{
						referenceId <- substr( element$text[[i]], nchar( .REFERENCE.STRING ) + 1, nchar( element$text[[i]] ) );
						referenceText <- .getReference( referenceId, report );
						
						if ( is.null( referenceText ) )
						{
							stop( paste( "Element", referenceId, "is referenced but not included in the report." ) );
						}
						
						updatedText <- paste( updatedText, referenceText, sep="" );
					}
					else
					{
						updatedText <-paste( updatedText, element$text[[i]], sep="" );
					}								
				}			
			}
			
			element$text <- updatedText;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			element$elements[[i]] <- .updateReferences( element$elements[[i]], report );
		}
	}		

	return ( element );
}


.updateResults <- function( element, report )
{
	# keeps track of all results that are encountered in this element
	resultList <- list();

	if ( !is.null( element$text ) )
	{
		if ( is.list( element$text ) )
		{
			updatedList <- list();
			j <- 1;
			
			if ( length( element$text ) > 0 )
			{
				for ( i in 1:length( element$text ) )
				{
					if ( is.list( element$text[[i]] ) )
					{
						# TODO: check if this is really of type result!!!!
						result <- element$text[[i]];
						
						if ( result$isSignificant == TRUE )
						{					
							if ( length( result$elements ) == 0 )
							{								
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "significant", "noevidence" ) );
							}
							else
							{
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "significant" ) );
							}							
						}
						else
						{
							if ( length( result$elements ) == 0 )
							{
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "noevidence" ) );
							}
							else
							{
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result" ) );
							}
						}
						j <- j + 1;
						
						result <- .updateReferences( result, report );
						
						updatedList[[j]] <- result$title;
						j <- j + 1;
						
						updatedList[[j]] <- .tag( "/span" );
						j <- j + 1;	
						
						# TODO: check if this is correct. the result should be added to the parent
						# add the result linked from the result summary to this element:
						#element$elements[[length( element$elements ) + 1]] <- result;
						resultList[[length( resultList ) + 1]] <- result;
					}
					else
					{
						updatedList[[j]] <-element$text[[i]];
						j <- j + 1;
					}				
				}			
			}
			
			element$text <- updatedList;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		updatedElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			update <- .updateResults( element$elements[[i]], report );
			
			updatedElements[[length( updatedElements ) + 1]] <- update$element;
			
			# if results were found in the element they will be inserted after the element into the report tree
			if ( length( update$results ) > 0 )
			{
				for ( r in 1:length( update$results ) )
				{
					updatedElements[[length( updatedElements ) + 1]] <- update$results[[r]];
				}
			}
		}
		
		# replace old elements with updated list
		element$elements <- updatedElements;
	}		

	return ( list( element=element, results=resultList ) );
}


.log <- function( ..., severity=.SEVERITY.LOG )
{			
	message <- .concat( ... );
	
	if ( severity == .SEVERITY.LOG )
	{
		message( "Log: ", message, "\n" );
	}
	
	if ( severity == .SEVERITY.WARNING )
	{
		warning( message );
	}
	
	if ( severity == .SEVERITY.ERROR )
	{
		stop( message );
	}
}


# --- element construction -------------------------------------------------------------------------

.newElement <- function( elementType, title=NULL, text=NULL, niceType=NULL, domId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- list( type = elementType );
	element$elements <- list();
	element$protection <- protection;
	
	element$id <- .getNextElementId();

	if ( !missing( title ) )
	{
		element$title <- title;
	}
	
	if ( !missing( text ) )
	{	
		element$text <- text;
	}
	
	if ( !missing( niceType ) )
	{	
		element$niceType <- niceType;
	}
	
	if ( !missing( domId ) )
	{
		element$domId <- domId;
	}
	
	return ( element );
}


newReport <- function( ..., version=0 )
{
	if ( !missing( ... ) )
	{
		element <- .newGeneralReport( ..., version=version );
	}
	else
	{
		element <- .newGeneralReport( "Report", version=version );
	}
	
	overview <- newSection( "Overview" );
	overview$domId <- "overview";

	introduction <- newSubSection( "Introduction" );
	introduction$domId <- "introduction";

	summary <- newSubSection( "Summary" );
	summary$domId  <- "summary";

	results <- newSection( "Results" );
	results$domId  <- "results";

	methods <- newSection( "Methods & Data" );
	methods$domId <- "methods";	

	input <- newSubSection( "Input" );
	input$domId <- "input";

	references <- newSubSection( "References" )
	references$domId <- "references";
	
	element$predefined <- list();
	element$predefined$overviewId <- overview$id;
	element$predefined$introductionId <- introduction$id;
	element$predefined$summaryId <- summary$id;
	element$predefined$resultsId <- results$id;
	element$predefined$methodsId <- methods$id;
	element$predefined$inputId <- input$id;
	element$predefined$referencesId <- references$id;
	
	element <- addTo( parent=element, addTo( parent=overview, introduction, summary ) );
	element <- addTo( parent=element, results );
	element <- addTo( parent=element, addTo( parent=methods, input, references ) );
}


.newGeneralReport <- function( ..., version=0 )
{
	element <- .newElement( .ELEMENT.REPORT, .concat( ... ) );
	
	element$meta <- list();
	
	# report version
	element$meta$version <- version;
	
	# report creator software
	element$meta$creator <- .getPackageVersion();
	
	# report creation date
	element$meta$createDate <- date();		
	
	# maintainer details
	element$meta$maintainer$name <- NA;
	element$meta$maintainer$email <- NA;
	element$meta$maintainer$affiliation <- NA;	
		
	# report navigation: parent, prev, next
	element$navigation <- list();
	
	element$navigation$parentUrl <- NA;
	element$navigation$parentName <- NA;

	element$navigation$nextUrl <- NA;
	element$navigation$nextName <- NA;

	element$navigation$previousUrl <- NA;
	element$navigation$previousName <- NA;


	return ( element );
}

# NOTE: This interface is experimental.
setNextReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$nextUrl <- url;
	report$navigation$nextName <- .concat( ... );	
	
	return ( report );
}

# NOTE: This interface is experimental.
setPreviousReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$previousUrl <- url;
	report$navigation$previousName <- .concat( ... );	
	
	return ( report );
}

# NOTE: This interface is experimental.
setParentReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$parentUrl <- url;
	report$navigation$parentName <- .concat( ... );	
	
	return ( report );
}


setMaintainerName <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$name <- .concat( ... );	

	return ( report );
}

getMaintainerName <- function( report )
{
	if ( is.null( report$meta$maintainer$name ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$name );
}

setMaintainerEmail <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$email <- .concat( ... );	

	return ( report );
}


getMaintainerEmail <- function( report )
{
	if ( is.null( report$meta$maintainer$email ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$email );
}

setMaintainerAffiliation <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$affiliation <- .concat( ... );	

	return ( report );
}

getMaintainerAffiliation <- function( report )
{
	if ( is.null( report$meta$maintainer$affiliation ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$affiliation );
}

setSignificantEntity <- function( report, ... )
{
	report$meta$significantEntity <- .concat( ... );

	return ( report );
}

getSignificantEntity <- function( report )
{
	if ( is.null( report$meta$significantEntity ) )
	{
		return ( DEFAULT.SIGNIFICANT.ENTITIY );
	}

	return ( report$meta$significantEntity );
}



.addToElementById <- function( element, id, ... )
{	
	if ( element$id == id )
	{
		element <- addTo( parent=element, ... );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			element$elements[[i]] <- .addToElementById( element$elements[[i]], id, ... );
		}
	}
	
	return ( element );	
}

.getElementById <- function( element, id )
{	
	if ( element$id == id )
	{
		return ( element );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			result <- .getElementById( element$elements[[i]], id );
			
			if ( !is.null( result ) )
			{
				return ( result );
			}
		}
	}
	
	return ( NULL );	
}

# removes element the given id from the tree using breadth-first search
.removeElementById <- function( element, id )
{	
	if ( length( element$elements ) > 0 )
	{	
		newElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			if ( element$elements[[i]]$id != id )
			{
				newElements[[length( newElements ) + 1]] <- element$elements[[i]];								
			}
		}
		
		element$elements <- newElements;
		
		if ( length( element$elements ) > 0 )
		{			
			for ( i in 1:length( element$elements ) )
			{
				element$elements[[i]] <- .removeElementById( element$elements[[i]], id );
			}		
		}
	}
	
	return ( element );	
}


addToResults <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$resultsId, ... ) ); 
}

addToMethods <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$methodsId, ... ) ); 
}

addToIntroduction <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$introductionId, ... ) ); 
}

addToSummary <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$summaryId, ... ) ); 
}

addToInput <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$inputId, ... ) ); 
}

addToReferences <- function( report, ... )
{
	return ( .addToElementById( report, report$predefined$referencesId, ... ) ); 
}


newSection <- function( ..., protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SECTION, .concat( ... ), protection=protection ) );
}


newSubSection <- function( ..., protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SUBSECTION, .concat( ... ), protection=protection ) );
}


newSubSubSection <- function( ..., protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SUBSUBSECTION, .concat( ... ), protection=protection ) );
}


.writeSection <- function ( element, file, level="" )
{
	if ( is.null( element$domId ) )
	{	
		.write( .tag( "div", class=.concat( level, "section" ) ), file );
	}
	else
	{
		.write( .tag( "div", class=.concat( level, "section" ), id=element$domId ), file );
	}

	.write( .tag( "div", class="sectionheader" ), file );
	.write( element$title, file );
	.write( .tag( "/div" ), file );
	
	.write( .tag( "div", class="sectionbody" ), file );
	.writeElements( element, file );
	.write( .tag( "/div" ), file );

	.write( .tag( "/div" ), file );	
}


newList <- function( ..., isNumbered=FALSE, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.LIST, protection=protection );
	
	element$isNumbered <- isNumbered;
	
	if ( missing( ... ) )
	{
		return ( element );
	}	
	
	args <- list( ... );
	
	# TODO: run the same checks in "addTo"
	for ( i in 1:length( args ) )
	{
		if ( args[[i]]$type == .ELEMENT.PARAGRAPH ||
			args[[i]]$type == .ELEMENT.LIST ||
			args[[i]]$type == .ELEMENT.RESULT )
		{
			element$elements[[length( element$elements ) + 1]] <- args[[i]];
		}
		else
		{
			stop( "Report elements of type ", args[[i]]$element$type, " can not be added to a list." );
		}
	}
	
	return ( element );
}

.writeList <- function( element, file )
{
	if ( element$isNumbered == TRUE )
	{
		.write( .tag( "ol" ), file );	
	}
	else
	{
		.write( .tag( "ul" ), file );			
	}
	
	for ( i in 1:length( element$elements ) )
	{
		if ( ( element$elements[[i]]$type != .ELEMENT.LIST ) && ( element$elements[[i]]$type != .ELEMENT.RESULT ) )
		{
			.write( .tag( "li" ), file );		
			.writeElement( element$elements[[i]], file );		
			.write( .tag( "/li" ), file );
		}
		else
		{		
			.writeElement( element$elements[[i]], file );		
		}
	}

	if ( element$isNumbered == TRUE )
	{
		.write( .tag( "/ol" ), file );	
	}	
	else
	{
		.write( .tag( "/ul" ), file );			
	}
}


newFigure <- function( file, ..., fileHighRes=NA, type=IMAGE.TYPE.RASTER, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.FIGURE, niceType="Figure", protection=protection );
	
	element$imageFilename <- file;
	element$imageFilenameHighRes <- fileHighRes;
	element$imageType <- type;
	
	element$text <- list( ... );
	
	return ( element );
}


.writeFigure <- function ( element, file )
{
	figureNumber <- element$number;
	
	# supplementary figure?
	if ( figureNumber < 0 )
	{
		figureNumber <- .concat( "S", abs( figureNumber ) );
	}
	
	.write( .tag( "div", class="figure", id=.concat( "figure_", figureNumber ) ), file );

	.write(	.tag( "div", class="caption" ), file );
	.write( .tag( "p" ), file );
	
	# "Figure 1. " ...
	.write( asStrong( element$niceType, "&nbsp;", figureNumber, ".&nbsp;" ), file );
	
	if ( !is.na( element$imageFilenameHighRes ) )
	{
		.write( "\n", .tag( "a", class="download", other=.concat( "href=\"", element$imageFilenameHighRes , "\"" ) ), "Get High-res Image", .tag( "/a" ), "\n", file );	
	}
	.write( element$text, file );
	.write( .tag( "/p" ), file );
	.write( .tag( "/div" ), file ); # caption

	.write( .tag( "div", class="image" ), file );
	
	
	
	if ( ( element$imageType == IMAGE.TYPE.RASTER ) || ( is.null( element$imageType ) ) )
	{
		.write( .tag( "img", other=.concat( "src=\"", element$imageFilename, "\"/" ) ), file );
	}
	else if ( element$imageType == IMAGE.TYPE.SVG )
	{
		.write( .tag( "object", other=.concat( "data=\"", element$imageFilename, "\" type=\"image/svg+xml\"" ) ), file );
		.write( .tag( "/object" ), file );
	}
	else
	{
		error( "Unknown image type selected in newFigure()." );
	}		
	
	.write( .tag( "/div" ), file ); # image
																
	.write( .tag( "/div" ), file ); # figure
}


newTable <- function( table, ..., file=NA, significantDigits=TABLE.SIGNIFICANT.DIGITS, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.TABLE, niceType="Table", protection=protection );
	
	element$table <- table;
	element$tableSignificantDigits <- significantDigits;
	element$tableFilename <- file;
	
	element$resultList <- list();
	element$resultIndices <- array( dim=dim( table ) );
	
	element$text <- list( ... );
	
	return ( element );
}


.trimTable <- function( table, significantDigits=3 )
{
	if ( significantDigits <= 0 )
	{
		return ( table );
	}
	
	for( c in 1:dim( table )[2] )
	{		
		# if this column is numeric ...
		if ( is.numeric( table[,c] ) == TRUE )
		{			
			# if there are entries with decimal digits ...
			if ( any( trunc( table[,c] ) != table[,c] ) )
			{
				# trim to significant digits
				table[,c] <- signif( table[,c], digits=significantDigits );
			}
			else
			{
				# add separators between every block of 3 digits
				# PROBLEM: TableSorter doesn't sort these numbers correctly
				#table[,c] <- format( table[,c], big.mark=",", scientific=FALSE );
			}
		}
	}
	
	return ( table );
}


.writeTable <- function ( element, file )
{	
	tableNumber <- element$number;
	
	# supplementary table?
	if ( tableNumber < 0 )
	{
		tableNumber <- .concat( "S", abs( tableNumber ) );
	}

	.write( .tag( "div", class="table", id=.concat( "table_", tableNumber ) ), file );
							
	.write(	.tag( "div", class="caption" ), file );
	.write( .tag( "p" ), file );
	
	# "Table 1. " ...
	.write( asStrong( element$niceType, "&nbsp;", tableNumber, ".&nbsp;" ), file );

	if ( !is.na( element$tableFilename ) )
	{
		.write( "\n", .tag( "a", class="download", other=.concat( "href=\"", element$tableFilename, "\"" ) ), "Get Full Table", .tag( "/a" ), "\n", file );	
	}

	.write( element$text, file );
	.write( .tag( "/p" ), file );
	.write( .tag( "/div" ), file ); # caption

	# write table only if a matrix/data frame has been supplied
	if ( is.null( dim( element$table ) ) )	
	{
		.write( .tag( "table", class="resulttable tablesorter sortabletable" ), file );
		
		# --- table header ---
	
		.write( .tag( "thead" ), file );
		.write( .tag( "tr" ), file );
		
		# write column names
		for( c in 1:length( element$table ) )
		{
			.write( .tag( "th" ), file );
			.write( element$table[c], file );
			.write( .tag( "/th" ), file );
		}
			
		.write( .tag( "/tr" ), file );
		.write( .tag( "/thead" ), file );		
		
		.write( .tag( "tbody" ), file );
		.write( .tag( "tr" ), file );
		.write( .tag( "td", other=.concat( "colspan=\"", length( element$table ), "\" align=\"center\"" ) ), file );
		.write( .tag( "a", other=.concat( "href=\"", element$tableFilename, "\"" ) ), "Get Full Table", .tag( "/a" ), file );		
		.write( .tag( "/td" ), file );
		.write( .tag( "/tr" ), file );
		.write( .tag( "/tbody" ), file );
		
		.write( .tag( "/table" ), file ); # resultstable			
	}
	else
	{
		# TODO: make sure that the resulttable class is attached to these tables
		#.write( .tag( "table", id=.concat( "sortable_table_", element$number ), class="tablesorter" ), file );
		#.write( .tag( "table", id=.concat( "sortable_table_", element$number ), class="resulttable tablesorter" ), file );
		.write( .tag( "table", class="resulttable tablesorter sortabletable" ), file );
		
		# --- table header ---
	
		.write( .tag( "thead" ), file );
		.write( .tag( "tr" ), file );
		
		# write column names
		for( c in 1:dim( element$table )[2] )
		{
			.write( .tag( "th" ), file );
			.write( colnames( element$table )[c], file );
			.write( .tag( "/th" ), file );
		}
			
		.write( .tag( "/tr" ), file );
		.write( .tag( "/thead" ), file );
		
		
		# --- table body ---
		
		# copy original table and perform trimming to significant digits	
		trimmedTable <- 0;
			
		if ( is.null( ( element$tableSignificantDigits ) ) )
		{
			trimmedTable <- .trimTable( element$table, TABLE.SIGNIFICANT.DIGITS );
		}
		else
		{	
			if ( is.na( element$tableSignificantDigits ) )
			{	
				trimmedTable <- .trimTable( element$table, TABLE.SIGNIFICANT.DIGITS );
			}
			else
			{
				trimmedTable <- .trimTable( element$table, element$tableSignificantDigits );
			}
		}
		
		.write( .tag( "tbody" ), file );
		
		for( r in 1:dim( element$table )[1] )
		{
			.write( .tag( "tr" ), file );
			
			for( c in 1:dim( element$table )[2] )
			{
				.write( .tag( "td" ), file );
				
				resultIndex <- element$resultIndices[r,c];
				
				if ( !is.na( resultIndex ) )
				{
					result <- element$resultList[[resultIndex]];
					
					if ( result$isSignificant )
					{					
						if ( length( result$elements ) == 0 )
						{
							.write( .tag( "span", class=c( "result", "significant", "noevidence" ), id=.concat( "resultid_", result$resultId ) ), file );
						}
						else
						{	
							.write( .tag( "span", class=c( "result", "significant" ), id=.concat( "resultid_", result$resultId ) ), file );						
						}
					}
					else
					{
						if ( length( result$elements ) == 0 )
						{
							.write( .tag( "span", class=c( "result", "noevidence" ), id=.concat( "resultid_", result$resultId ) ), file );
						}
						else
						{
							.write( .tag( "span", class=c( "result" ), id=.concat( "resultid_", result$resultId ) ), file );
						}
					}
				}
				
				#.write( as.character( element$table[r,c] ), file );
				#.write( as.character( trimmedTable[r,c] ), file );
				.write( format( trimmedTable[r,c], scientific=1 ), file );
				
				if ( !is.na( resultIndex ) )
				{
					.write( .tag( "/span" ), file );
				}
	
				.write( .tag( "/td" ), file );						
			}		
			
			.write( .tag( "/tr" ), file );
		}
			
		.write( .tag( "/tbody" ), file );
		
		.write( .tag( "/table" ), file ); # resultstable	
	}
	
	.write( .tag( "/div" ), file ); # table	

	# write all elements, i.e. results only in this case
	.writeElements( element, file );																		
}



newCitation <- function( authors="", title, publication="", issue="", number="", pages="", year="", url="" )
{
	element <- .newElement( .ELEMENT.CITATION, niceType="Citation" );
	
	element$authors <- authors;
	element$year <- year;
	element$title <- title;
	element$publication <- publication;
	element$publicationIssue <- issue;
	element$publicationNumber <- number;
	element$pages <- pages;
	element$url <- url;	
	
	return ( element );
}

newJournalCitation <- function( authors, title, publication, issue, number, pages, year, url="" )
{
	return ( newCitation( authors=authors, title=title, publication=publication, issue=issue, number=number, pages=pages, year=year, url=url ) );
}

newWebCitation <- function( authors, title, url )
{
	return ( newCitation( authors=authors, title=title, url=url ) );
}


.writeCitation <- function( element, file )
{
	.write( .tag( "div", class="reference", id=.concat( "reference_", element$number ) ), file );
	
	.write( "[", element$number, "] ", nobreak=TRUE, file );

	if ( element$authors != "" )
	{
		.write( element$authors, ", ", nobreak=TRUE, file );
	}

	if ( element$title != "" ) # only required field
	{
		.write( asLink( url=element$url, element$title ), nobreak=TRUE, file );
	}

	if ( element$publication != "" )
	{
		.write( ", ", asEmph( element$publication ), nobreak=TRUE, file );
	}

	if ( element$publication != "" && element$publicationIssue != "" )
	{
		.write( " ", asStrong( element$publicationIssue ), nobreak=TRUE, file );
	}

	if ( element$publication != "" && element$publicationIssue != "" && element$publicationNumber != "" )
	{
		.write( "(", element$publicationNumber, ")", nobreak=TRUE, file );
	}

	if ( element$pages != "" )
	{
		.write( ":", element$pages, nobreak=TRUE, file );
	}
	
	if ( element$year != "" )
	{
		.write( " (",element$year, ")", nobreak=TRUE, file );
	}
	
	.write( .tag( "/div" ), file ); # reference	
}


newParagraph <- function( ..., protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.PARAGRAPH, protection=protection );

	element$text <- list( ... );
	
	return ( element );
}

.writeParagraph <- function ( element, file )
{
	.write( .tag( "p" ), file );
	.write( element$text, file );
	
	# write all elements, i.e. results only in this case
	.writeElements( element, file );	
	
	.write( .tag( "/p" ), file );	
}


# does not need a write function, writing is provided by .writeList
newParameterList <- function( ..., separator=" = ", protection=PROTECTION.PUBLIC )
{
	element <- newList( isNumbered=FALSE, protection=protection );
	
	args <- list( ... );
	
	if ( length( args ) < 2 )
	{
		stop( "At least one parameter-value pair has to be provided to parameter list." );
	}
	
	if ( length( args ) %% 2 != 0 )
	{
		stop( "Number of elements in parameter list is not multiple of 2." );
	}
	
	j <- 1;
	
	for ( i in 1:(length( args )/2) )
	{
		element <- addTo( parent=element, newParagraph( args[[j]], separator, args[[j+1]] ) );
		
		j <- j + 2;
	}
	
	return ( element );
}


.writeParagraph <- function ( element, file )
{
	.write( .tag( "p" ), file );
	.write( element$text, file );
	
	# write all elements, i.e. results only in this case
	.writeElements( element, file );	
	
	.write( .tag( "/p" ), file );	
}


newResult <- function( ..., isSignificant=FALSE, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.RESULT, .concat( ... ), protection=protection );

	element$resultId <- .getNextResultId();
	element$isSignificant <- isSignificant;
	
	return ( element );
}


.writeResult <- function ( element, file )
{
	.write( .tag( "div", class="evidence", id=.concat( "evidenceid_", element$resultId ) ), file );

	.writeElements( element, file );

	.write( .tag( "/div" ), file );	
}


addTo <- function( parent, ..., row=NA, column=NA )
{
	args <- list( ... );
	
	if ( parent$type == .ELEMENT.TABLE )
	{
		if ( !missing( row ) && !missing( column ) )
		{			
			if ( length( args ) == 1 )
			{
				if ( args[[1]]$type == .ELEMENT.RESULT )
				{
					parent <- .addResult( table=parent, row=row, column=column, result=args[[1]] );
					
					return ( parent );
				}
			}
		}
	}
	
	
	if ( length( args ) > 0 )
	{
		for ( i in 1:length( args ) )
		{
			parent$elements[[length( parent$elements ) + 1]] <- args[[i]];	
		}	
	}
		
	return ( parent );
}


.addResult <- function( table, row, column, result )
{
	# !!! do not change the order of the following two lines!!!
	table$resultList[[length( table$resultList ) + 1]] <- result;	
	table$resultIndices[row,column] = length( table$resultList );
	
	# add result right after table
	table$elements[[length( table$elements ) + 1]] <- result;	
	
	return ( table );
}



asReference <- function( element )
{
	# TODO: create link if requested
	# TODO: check for type!
	return ( paste( .REFERENCE.STRING, element$id, sep="" ) );
}

asSummary <- function( result )
{
	return ( result );
}


asLink <- function( url, ... )
{
	# TODO: open link in new window if requested		
	
	args <- list( ... );
	
	if ( missing( url ) && length( args ) == 0 )
	{
		stop( "Unable to create link without URL or link text." );
	}	
	
	if ( url == "" || missing( url ) )
	{
		return ( .concat( args ) );	# just print text but do not create a link
	}
	
	if ( length( args ) == 0 )
	{
		args <- url; # print URL as link text
	}
	
	# TODO: remove "\n"s
	#return ( .concat( "\n", .tag( "a", other=.concat( "href=\"", url, "\"" ) ), args, .tag( "/a" ), "\n" ) );
	return ( .concat( .tag( "a", other=.concat( "href=\"", url, "\"" ) ), args, .tag( "/a" ) ) );
}


#asIdentifier <- function( label, class=NA, column=NA, row=NA )
#{
#	if ( missing( class ) && missing( label ) )
#	{
#		stop( "Unable to create identifier without class or label." );
#	}	
#	
#	if ( missing( class ) )
#	{
#		return ( label ); # just print text
#	}
#	
#	return ( .concat( "<span class=\"identifier ", class, "\">", label, "</span>" ) );
#}


asStrong <- function( ... )
{
	return ( .concat( .tag( "strong" ), ..., .tag( "/strong" ) ) );
}


asEmph <- function( ... )
{
	return ( .concat( .tag( "em" ), ..., .tag( "/em" ) ) );
}


asParameter <- function( ... )
{
	return ( .concat( .tag( "span", class="parameter" ), ..., .tag( "/span" ) ) );
}


asValue <- function( ... )
{
	return ( .concat( .tag( "span", class="value" ), ..., .tag( "/span" ) ) );
}


asFilename <- function( ... )
{
	return ( .concat( .tag( "span", class="filename" ), ..., .tag( "/span" ) ) );
}


asCode <- function( ... )
{
	return ( .concat( .tag( "span", class="code" ), ..., .tag( "/span" ) ) );
}


.concat <- function( ... )
{
	if ( length( list( ... ) ) == 0 )
	{
		return ( "" );
	}
	
	return ( paste( ..., sep="" ) );
}


# helper function: writes a line of html code
.write <- function( ..., nobreak=FALSE, file )
{
  if ( missing( file ) )
  {
	args <- list( ... );
  	file <- args[[length( args )]];
  	args[[length( args )]] <- "";
  }  
  
  # string to be written to file
  string <- paste( args, sep="" );

  # if the report is supposed to be used as a fragment, all ids have to be made unique
  # scan text for this pattern "< ...id="xyz" ...>"  
  # replace with "< ...id="xyz_REPORTID" ...>"
  # sed s/"\(<.*\)id=\"\([A-Za-z][A-Za-z0-9_\.-\:]*\)\"\(.*>\)"/\\1id=\\2_TEXT\\3/g demo.html      
  if ( .getGlobalReportId() != "" )
  {
  	 string <- gsub( "(<.*)id=\"([A-Za-z][A-Za-z0-9_\\.-\\:]*)\"(.*>)", .concat( "\\1id=\"\\2_", .getGlobalReportId(), "\"\\3" ), string );  
  }	 
  
  if ( nobreak )
  {  
  	cat( string, file=file, sep="" );
  }
  else
  {
  	cat( string, "\n", file=file, sep="" );
  }
}


.tag <- function( tag, id=NA, class=NA, other=NA )
{
	idString <- "";
	classString <- "";
	otherString <- "";
	
	if ( !missing( id ) )
	{
		idString <- .concat( " id=\"", id, "\"" );	
	}

	if ( !missing( class ) )
	{
		classString <- .concat( " class=\"", paste( class, sep=" ", collapse=" " ), "\"" );	
	}
	
	if ( !missing( other ) )
	{
		otherString <- .concat( " ", other );
	}
	
	return ( .concat( "<", tag, classString, idString, otherString, ">" ) );
}

.comment <- function( ... )
{
	return ( .concat( "<!-- ", ..., " -->" ) );
}



.writeElement <- function( element, file )
{
	if ( element$type == .ELEMENT.PARAGRAPH )
	{
		.writeParagraph( element, file );
	}

	if ( element$type == .ELEMENT.SECTION )
	{
		.writeSection( element, file, "" );
	}

	if ( element$type == .ELEMENT.SUBSECTION )
	{
		.writeSection( element, file, "sub" );
	}

	if ( element$type == .ELEMENT.SUBSUBSECTION )
	{
		.writeSection( element, file, "subsub" );
	}

	if ( element$type == .ELEMENT.LIST )
	{
		.writeList( element, file );
	}

	if ( element$type == .ELEMENT.FIGURE )
	{
		.writeFigure( element, file );
	}

	if ( element$type == .ELEMENT.TABLE )
	{
		.writeTable( element, file );
	}

	if ( element$type == .ELEMENT.CITATION )
	{
		.writeCitation( element, file );
	}

	if ( element$type == .ELEMENT.RESULT )
	{
		.writeResult( element, file );
	}
}


.writeElements <- function( element, file )
{
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			.writeElement( element$elements[[i]], file );
		}
	}
}


getSummary <- function( report )
{
	.initNumbers();

	return ( .updateReferences( .updateNumberedElements( .getElementById( report, report$predefined$summaryId ) ), report=report )$elements ); 	
}


getSignificantResultsCount <- function( report )
{
	return ( .countSignificantResults( report ) );
}


writeReport <- function( report, filename=DEFAULT.REPORT.FILENAME, debug=FALSE, output=c( HTML.REPORT, HTML.FRAGMENT, RDATA.REPORT ), level=PROTECTION.PUBLIC, debugCss=NA, debugJavaScript=NA )
{	
	# check if input and references contain any reports, if not, remove those sections
	references <- .getElementById( report, report$predefined$referencesId );
	input <- .getElementById( report, report$predefined$inputId );
	
	if ( !is.null( references ) )
	{		
		# remove reference section
		report <- .removeElementById( report, references$id );

		if ( length( references$elements ) > 0 )
		{
			# there are some references, insert them at the end of the methods section
			report <- addToMethods( report, references );
		}					
	}

	if ( !is.null( input ) )
	{
		if ( length( input$elements ) == 0 )
		{
			# no input, remove the empty section
			report <- .removeElementById( report, input$id );
		}
	}
	
	
	# --- additional meta information --- #
	
	# assign report id
	report$reportId <- .newReportId();
	
	# assign date of report writing (if doesn't exist yet)
	if ( is.null( report$meta$createDate ) )
	{
		report$meta$createDate <- date();
	}	

	# write as requested output types	
	if ( HTML.REPORT %in% output )
	{
		.writeGeneralReport( report, .concat( filename, "_standalone.html" ), debug, HTML.REPORT, FALSE, level=level, debugCss=debugCss, debugJavaScript=debugJavaScript );		
	}
	
	if ( HTML.FRAGMENT %in% output )
	{
		.writeGeneralReport( report, .concat( filename, ".html" ), debug, HTML.FRAGMENT, FALSE, level=level, debugCss=debugCss, debugJavaScript=debugJavaScript );
	}
	
	if ( RDATA.REPORT %in% output )
	{		
		save( report, file=.concat( filename, ".RData" ) );		
	}
}


.writeGeneralReport <- function( report, filename, debug=FALSE, output=HTML.REPORT, append=FALSE, level=PROTECTION.PUBLIC, debugCss=NA, debugJavaScript=NA )
{
	# initialize counters for all numbered objects (figures, tables, references, etc.)
	.initNumbers();
							
	report <- .removeProtectedElements( report, level );
	
	# extract result elements (e.g. from paragraphs) and insert into the report tree
	report <- .updateResults( report, report=report )$element; # if there are results directly in the report they will be ignored
	report <- .updateNumberedElements( report );
	report <- .updateReferences( report, report=report );
	
	file <- NA;
	
	if ( append )
	{
		file <- file( filename, "a" );
	}
	else
	{
		file <- file( filename, "w" );
	}
	
	# write file header if in debug mode or if writing standalone report
	if ( output == HTML.REPORT )
	{
		.write( "<!DOCTYPE html>", file );
		
		.write( .tag( "html" ), file );
		
		.write( .tag( "head" ), file );
		
		# write report title
		.write( .tag( "title" ), file );
		.write( report$title, file );
		.write( .tag( "/title" ), file );					
		
		.write( .tag( "/head" ), file );
		
		.write( .tag( "body" ), file );
	}	
	
	.write( .tag( "div", class="report", id=report$reportId ), file );
	
	.write( .comment( "Nozzle Report Id = ", report$reportId ), file );
	.write( .comment( "Nozzle Report Version = ", report$meta$version ), file );

	.write( .comment( "Nozzle Report Create Package = ", report$meta$creator ), file );
	.write( .comment( "Nozzle Report Create Date = ", report$meta$createDate ), file );
	
	report$meta$renderer <- .getPackageVersion();
	report$meta$renderDate <- date();

	.write( .comment( "Nozzle Report Render Package = ", report$meta$renderer ), file );
	.write( .comment( "Nozzle Report Render Date = ", report$meta$renderDate ), file );
	
	# add further meta information
	report$meta$reportTitle <- report$title;
	report$meta$reportId <- report$reportId;
	
		
	# insert Nozzle CSS
	if ( debug )
	{
		if ( is.na( debugCss ) )
		{
			.write( .tag( "style" ), file );	
			.write( paste( get( "cssCode", env=.nozzleEnv ), collapse="\n" ), file );
			.write( .tag( "/style" ), file );
		}
		else
		{
			.write( "<link type=\"text/css\" rel=\"stylesheet\" href=\"", debugCss, "\">", file );
		}
	}
	else
	{
		.write( .tag( "style" ), file );	
		.write( paste( get( "cssMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( .tag( "/style" ), file );
	}
	
	# embedded jQuery JavaScript (only in standalone)
	if ( output == HTML.REPORT )
	{
		#.write( "<script type=\"text/javascript\" src=\"https://ajax.googleapis.com/ajax/libs/jquery/1.5.0/jquery.min.js\"></script>", file );
		.write( "<script type=\"text/javascript\">", file );
		.write( paste( get( "jQueryMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( "</script>", file );	
	}
	
	# embedded TableSorter JavaScript
	.write( "<script type=\"text/javascript\">", file );
	if ( debug )
	{
		.write( paste( get( "jQueryTableSorterMinCode", env=.nozzleEnv ), collapse="\n" ), file );
	}
	else
	{
		.write( paste( get( "jQueryTableSorterMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
	}
	.write( "</script>", file );	
	

	# embedded Nozzle JavaScript
	if ( debug )
	{
		if ( is.na( debugJavaScript ) )
		{
			.write( "<script type=\"text/javascript\">", file );
			.write( paste( get( "javaScriptCode", env=.nozzleEnv ), collapse="\n" ), file );
			.write( "</script>", file );	
		}
		else
		{
			.write( "<script type=\"text/javascript\" src=\"", debugJavaScript ,"\"></script>", file );			
		}
	}
	else
	{
		.write( "<script type=\"text/javascript\">", file );
		.write( paste( get( "javaScriptMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( "</script>", file );	
	}


	if ( ( output == HTML.REPORT ) || ( append == TRUE ) )
	{
		.write( "<script type=\"text/javascript\">jQuery(document).ready( function() { initNozzle( \"", report$reportId, "\", $ ); });</script>", file );
	}
	else if ( output == HTML.FRAGMENT )
	{
		.write( "<script type=\"text/javascript\">function onReportLoad_", report$reportId, "() { initNozzle( \"", report$reportId, "\", $j ); }</script>", file );		
	}
	else
	{
		# warning!
	}
	
	
	# === REPORT META BEGIN ===	
	
	.write( "<script type=\"text/javascript\">", file );
	.writeJsonElement( "nozzleMeta", report$meta, file );
	.writeJsonElement( "nozzleNavigation", report$navigation, file );
	.write( "</script>", file );	
	
	# === REPORT META END ===	


	# === REPORT CONTENT BEGIN ====

	.write( .tag( "div", class="frame" ), file );	
	.write( .tag( "div", class="main" ), file );
		
	# set global report id if this is being written as a fragment
	# if set to anything else but "", .write will attach this to all ids in HTML tags
	.setGlobalReportId( report$reportId );
			
	# write title
	.write( .tag( "div", class="title" ), report$title, .tag( "/div" ), file );
	
	# write maintainer info
	if ( !is.na( getMaintainerName( report ) ) )
	{
		.write( .tag( "div", class="maintainer" ), file );
		
		if ( !is.na( getMaintainerEmail( report ) ) )
		{
			.write( "Maintained by ", .tag( "span", class="name" ), asLink( url=.concat( "mailto:", getMaintainerEmail( report ) ), getMaintainerName( report ) ), .tag( "/span" ), file );
		}
		else
		{
			.write( "Maintained by ", .tag( "span", class="name" ), getMaintainerName( report ), .tag( "/span" ), file );
		}

		if ( !is.na( getMaintainerAffiliation( report ) ) )
		{
			.write( " (", .tag( "span", class="affiliation" ), getMaintainerAffiliation( report ), .tag( "/span" ), ")", file );
		}
				
		.write( .tag( "/div" ), file );
	}
	
	# write content
	for ( i in 1:length( report$elements ) )
	{
		.writeElement( report$elements[[i]], file );		
	}

	.write( .tag( "/div" ), file ); # main
	.write( .tag( "/div" ), file ); # frame				
	.write( .tag( "/div" ), file ); # report

	# === REPORT CONTENT END ====

	
	# write file footer if in debug mode or standalone report
	if ( output == HTML.REPORT )
	{	
		.write( .tag( "/body" ), file );
		
		.write( .tag( "/html" ), file );	
	}


	# reset global report id if this is being written as a fragment to avoid conflicts with other reports
	.setGlobalReportId( "" );
	
	close( file );
}


# writes a meta entry into the header into file unless the value is NA with write.na == FALSE
.writeMetaInformation <- function( attribute, value, file, write.na=FALSE )
{
	if ( !is.na( value ) || write.na )
	{
		.write( .tag( "meta", other=.concat( "attr=\"", attribute, "\" value=\"", value, "\"" ) ), file );
		.write( .tag( "/meta" ), file );
	}
}


# TODO: remove this test function
.testJson <- function( filename )
{
	file <- file( filename, "w" );
	
	l <- list( z=NA, a=50, b="Nils", c=c( 1,2,3 ), d=c("a","b","c"), e=list( one="One", two="2", three=c("i", "i", "i") ), f="eff", g=list( one="Eins", two="Zwee", three=c(3, "i", "i") ), h="HAH" );
	
	.writeJsonElement( "test", l, file );
	
	close( file );
}


.writeJsonElement <- function( jsVariable, element, file )
{
	.write( "var ", jsVariable, " = {", file );
	
	.writeJsonList( NA, element, file, indent=3 );
		
	.write( "};", file );		
}


.writeJsonNumeric <- function( name, value, file, indent=0 )
{
	if ( length( value ) == 1 )
	{
		.write( .createIndent( indent ), "\"", name, "\": ", value, file, nobreak=T );
	}
	else
	{
		.write( .createIndent( indent ), "\"", name, "\": [", paste( value, collapse=", " ), "]", file, nobreak=T );	
	}
}


.writeJsonCharacter <- function( name, value, file, indent=0 )
{
	if ( length( value ) == 1 )
	{
		.write( .createIndent( indent ), "\"", name, "\": ", "\"", value, "\"", file, nobreak=T );
	}
	else
	{
		.write( .createIndent( indent ), "\"", name, "\": [", paste( paste( "\"", value, "\"", sep="" ), collapse=", " ), "]", file, nobreak=T );	
	}
}


.writeJsonNull <- function( name, file, indent=0 )
{
	.write( .createIndent( indent ), "\"", name, "\": ", "null", file, nobreak=T );
}


.writeJsonList <- function( name, element, file, indent=0 )
{
	if ( !is.na( name ) )
	{
		.write( .createIndent( indent ), "\"", name, "\": {", file );
	}

	# get names	
	listNames <- names( element );
		
	# iterate over names and write out values
	for ( i in 1:length( listNames ) )
	{
		value <- element[[listNames[i]]];
		
		if ( is.list( value ) )
		{
			.writeJsonList( listNames[i], value, file, indent=indent+3 ); 
		}
		else if ( is.na( value ) )
		{
			.writeJsonNull( listNames[i], file, indent=indent+3 ); 
		}
		else if ( is.numeric( value ) )
		{
			.writeJsonNumeric( listNames[i], value, file, indent=indent+3 );
		}
		else		
		{
			.writeJsonCharacter( listNames[i], value, file, indent=indent+3 );		
		}
		
		if ( i < length( listNames ) )
		{
			.write( ",", file );
		}
		else
		{
			.write( "\n", file, nobreak=T );
		}
	}
	
	# closing braces
	if ( !is.na( name ) )
	{
		.write( .createIndent( indent ), "}", file, nobreak=T );
	}
}

.createIndent <- function( indent, character=" " )
{
	return( paste( rep( character, indent ), collapse="" ) );
}
